#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/8/4
 """
from __future__ import unicode_literals, absolute_import
from django.conf import settings
from django.utils.encoding import force_unicode
from djdg_core.utils.hailstone import regular_url
from djdg_core.exceptions import StanderResponseError
from djdg_oauth.oauthcore import getSign
import requests
from requests.models import complexjson
from requests_toolbelt import MultipartEncoder


class RestClient(object):
    """
    restful 接口调用客户端
    """
    default_headers = (
        ('Content-Type', 'application/json;charset=utf-8'),
    )

    def __init__(self,
                 host,
                 base_url='/',
                 protocol='http',
                 headers=None,
                 logger=None,
                 oauth_app_id=None,
                 oauth_app_secret=None
                 ):
        self.host = regular_url(host)
        self.base_url = regular_url(base_url)
        self.protocol = protocol
        headers = headers if headers else ()
        self.__headers = headers + self.default_headers
        self.logger = logger
        self.oauth_app_id = oauth_app_id
        self.oauth_app_secret = oauth_app_secret

    def get_headers(self):
        d = {}
        for k, v in self.__headers:
            d[k] = v
        return d

    def get_resource(self, name, app=None, token=None):
        path = regular_url(name)
        if app:
            path = '{0}{1}'.format(regular_url(app), path)
        url = '{protocol}://{host}{base_url}{path}'.format(
            protocol=self.protocol,
            base_url=self.base_url,
            host=self.host,
            path=path
        )
        return RestResource(url, self, token)


class RestResource(object):
    """
    restful 资源
    """
    methods = ('get', 'post', 'put', 'delete', 'options', 'head', 'patch')

    def __init__(self, url, client, token=None):
        self.url = url
        self.client = client
        self.token = token
        self.logger = getattr(client, 'logger')

    def do_request(self, method, action=None, data=None, params=None, headers=None, json=None, raise_exception=True,
                   end_slash=True, multi_part_fields=None):
        send_headers = self.client.get_headers()
        if self.token:
            send_headers['token'] = self.token
        if headers:
            send_headers.update(headers)
        if self.client.oauth_app_id is not None and self.client.oauth_app_secret is not None:
            # 认证中间件先检查有误Authorization，后检查有无token
            # send_headers.pop('token', None)
            if 'get' == method:
                params = params if params else {}
                params['appid'] = self.client.oauth_app_id
                sign_body = params
                send_headers.pop('Content-Type', None)
            elif multi_part_fields:
                m = MultipartEncoder(fields=multi_part_fields)
                data = m.to_string()
                send_headers['Content-Type'] = m.content_type
                sign_body = data
            else:
                json = json if json else {}
                json['appid'] = self.client.oauth_app_id
                data = complexjson.dumps(json)
                sign_body = data
            send_headers['Authorization'] = getSign(sign_body, self.client.oauth_app_secret)
            send_headers['appid'] = self.client.oauth_app_id
        if action:
            action = regular_url(action)
            url = '{0}{1}'.format(self.url, action)
        else:
            url = self.url
        if not end_slash:
            url = url.rstrip('/')
        res = requests.request(method, url, data=data, params=params, headers=send_headers, json=json)
        ret = Result(res)
        if self.logger:
            content_type = send_headers.get('Content-Type') or ''
            content_type = content_type.lower()
            if content_type.startswith('multipart'):
                body = content_type
            else:
                body = force_unicode(res.request.body) if res.request.body is not None else None
            msg = 'request {uri} {method} {status_code} \nheaders: {headers}\nbody: {body}\nresponse: {resp}'.format(
                uri=res.request.url,
                method=res.request.method,
                status_code=res.status_code,
                headers='; '.join(['{k}: {v}'.format(k=k, v=v) for k, v in send_headers.items()]),
                body=body,
                resp=force_unicode(res.content)
            )
            self.logger.info(msg)
        if raise_exception and ret.status != settings.RESPONSE_CONFIG['OK_STATUS']:
            raise StanderResponseError(ret.status, ret.msg, getattr(ret, 'data', {}))
        return ret

    def __getattr__(self, item):
        method = item.lower()
        action = None
        if method not in self.methods:
            for m in self.methods:
                if method.startswith(m):
                    method = m
                    action = item[len(m) + 1:]
                    break
            else:
                action = item
        return Method(self, method, action)


class Method(object):
    """
    模拟请求
    """
    def __init__(self, resource, method=None, action=None):
        self.resource = resource
        self.method = method
        self.action = action
        self.token = None

    def __call__(self, data=None, params=None, method=None, headers=None, action_to=None, raise_exception=True,
                 end_slash=True, multi_part_fields=None, **kwargs):
        if method:
            self.method = method
        if not self.method:
            self.method = 'post' if data else 'get'
        if 'get' == self.method:
            params = params if params else {}
            params.update(kwargs)
        else:
            data = data if data else {}
            data.update(kwargs)
        action = ''
        if action_to:
            action = '{0}/'.format(action_to)
        if self.action:
            action = '{0}{1}/'.format(action, self.action)
        headers = headers if headers else {}
        if self.token:
            headers['token'] = self.token
        return self.resource.do_request(self.method,
                                        action=action,
                                        json=data,
                                        params=params,
                                        headers=headers,
                                        raise_exception=raise_exception,
                                        end_slash=end_slash,
                                        multi_part_fields=multi_part_fields
                                        )

    def __getattr__(self, item):
        method = item.lower()
        if method not in self.resource.methods:
            for m in self.resource.methods:
                if method.startswith(m):
                    self.method = m
                    self.action = item[len(m) + 1:]
                    break
            else:
                self.action = item
        else:
            self.method = method
        return self

    def set_token(self, token):
        self.token = token
        return self


class Result(object):
    """
    restful 响应数据
    """
    def __init__(self, response):
        self.__resp = response
        self.__data = None
        self.__result = None

    @property
    def result(self):
        if self.__result is None:
            if 500 == self.status_code:
                self.__result = {
                    settings.RESPONSE_CONFIG['MSG_PARAM']: self.__resp.text,
                    settings.RESPONSE_CONFIG['DATA_PARAM']: {},
                    settings.RESPONSE_CONFIG['STATUS_PARAM']: self.status_code
                }
            else:
                try:
                    self.__result = self.__resp.json()
                except Exception as e:
                    raise StanderResponseError(2, e.message)
        return self.__result

    @property
    def data(self):
        return self.result.get(settings.RESPONSE_CONFIG['DATA_PARAM'], None)

    def __getattr__(self, item):
        return self.result[item]

    @property
    def status(self):
        return self.result[settings.RESPONSE_CONFIG['STATUS_PARAM']]

    @property
    def status_code(self):
        return self.__resp.status_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2017/10/18
 """
from __future__ import unicode_literals, absolute_import
import logging
import string
import random
import time
from collections import namedtuple

import redis
from redis.exceptions import RedisError

# Python 3 compatibility
string_type = getattr(__builtins__, 'basestring', str)

try:
    basestring
except NameError:
    basestring = str


Lock = namedtuple("Lock", ("validity", "resource", "key"))


class CannotObtainLock(Exception):
    pass


class MultipleRedlockException(Exception):
    def __init__(self, errors, *args, **kwargs):
        super(MultipleRedlockException, self).__init__(*args, **kwargs)
        self.errors = errors

    def __str__(self):
        return ' :: '.join([str(e) for e in self.errors])

    def __repr__(self):
        return self.__str__()


class Redlock(object):

    default_retry_count = 3
    default_retry_delay = 0.2
    clock_drift_factor = 0.01
    unlock_script = """
    if redis.call("get",KEYS[1]) == ARGV[1] then
        return redis.call("del",KEYS[1])
    else
        return 0
    end"""

    def __init__(self, connection_list, retry_count=None, retry_delay=None):
        self.servers = []
        for connection_info in connection_list:
            try:
                if isinstance(connection_info, string_type):
                    server = redis.StrictRedis.from_url(connection_info)
                elif type(connection_info) == dict:
                    server = redis.StrictRedis(**connection_info)
                else:
                    server = connection_info
                self.servers.append(server)
            except Exception as e:
                raise Warning(str(e))
        self.quorum = (len(connection_list) // 2) + 1

        if len(self.servers) < self.quorum:
            raise CannotObtainLock(
                "Failed to connect to the majority of redis servers")
        self.retry_count = retry_count or self.default_retry_count
        self.retry_delay = retry_delay or self.default_retry_delay

    def lock_instance(self, server, resource, val, ttl):
        try:
            assert isinstance(ttl, int), 'ttl {} is not an integer'.format(ttl)
        except AssertionError as e:
            raise ValueError(str(e))
        return server.set(resource, val, nx=True, px=ttl)

    def unlock_instance(self, server, resource, val):
        try:
            # server.eval(self.unlock_script, 1, resource, val)
            r = server.get(resource)
            if val == r:
                server.delete(resource)
        except Exception as e:
            logging.exception("Error unlocking resource %s in server %s", resource, str(server))

    def get_unique_id(self):
        CHARACTERS = string.ascii_letters + string.digits
        return ''.join(random.choice(CHARACTERS) for _ in range(22)).encode()

    def lock(self, resource, ttl):
        retry = 0
        val = self.get_unique_id()

        # Add 2 milliseconds to the drift to account for Redis expires
        # precision, which is 1 millisecond, plus 1 millisecond min
        # drift for small TTLs.
        drift = int(ttl * self.clock_drift_factor) + 2

        redis_errors = list()
        while retry < self.retry_count:
            n = 0
            start_time = int(time.time() * 1000)
            del redis_errors[:]
            for server in self.servers:
                try:
                    if self.lock_instance(server, resource, val, ttl):
                        n += 1
                except RedisError as e:
                    redis_errors.append(e)
            elapsed_time = int(time.time() * 1000) - start_time
            validity = int(ttl - elapsed_time - drift)
            if validity > 0 and n >= self.quorum:
                if redis_errors:
                    raise MultipleRedlockException(redis_errors)
                return Lock(validity, resource, val)
            else:
                for server in self.servers:
                    try:
                        self.unlock_instance(server, resource, val)
                    except:
                        pass
                retry += 1
                time.sleep(self.retry_delay)
        return False

    def unlock(self, lock):
        redis_errors = []
        for server in self.servers:
            try:
                self.unlock_instance(server, lock.resource, lock.key)
            except RedisError as e:
                redis_errors.append(e)
        if redis_errors:
            raise MultipleRedlockException(redis_errors)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/8/9
 """
from __future__ import unicode_literals, absolute_import
from django.forms.fields import DateTimeField
import datetime


class TimestampField(DateTimeField):
    """
    时间戳字段类
    """
    def to_python(self, value):
        try:
            val = int(value)
            val = datetime.datetime.fromtimestamp(val)
            if val is not None:
                value = val
        except:
            pass
        return super(TimestampField, self).to_python(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/19
 """
from __future__ import unicode_literals, absolute_import
from django.conf.urls import include
from django.conf.urls import url
from rest_framework.routers import DefaultRouter
from djdg_core.common_api import views


sms_router = DefaultRouter()
sms_router.register('sms', views.SmsViewSet, base_name='sms')

register_router = DefaultRouter()
register_router.register('register', views.RegisterViewSet, base_name='register')

file_router = DefaultRouter()
file_router.register('files', views.FileUploadViewSet, base_name='files')


urlpatterns = [
    url('^common_api/', include(sms_router.urls), name='sms'),
    url('^common_api/', include(register_router.urls), name='register'),
    url('^common_api/', include(file_router.urls), name='files'),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2017/8/2
 """
from __future__ import unicode_literals, absolute_import
from celery.app import shared_task
from djdg_core.settings import LOGGER_NAME
import logging


class Subject(object):
    """
    可订阅的主题
    """
    def __init__(self, name, hub):
        self.name = name
        self.hub = hub
        self.hub.register_subject(self)

    def publish(self, *args, **kwargs):
        self.hub.publish(self, *args, **kwargs)

    def async_publish(self, app, *args, **kwargs):
        task_name = celery_task_of_hubs.get(self.hub)
        if not task_name:
            raise Exception('订阅发布总线{}找不到对应的celery任务名称！'.format(self.hub.__class__.name))
        _args = [self.name]
        _args.extend(args)
        app.send_task(
            task_name,
            _args,
            kwargs
        )


class Subscriber(object):
    """
    订阅者
    """
    def __init__(self, hub, subject, *args, **kwargs):
        self.hub = hub
        self.subject = subject
        self.args = args
        self.kwargs = kwargs
        context = kwargs.pop('context', None) or {}
        self.context = context

    def subscribe(self, subject, priority=20):
        self.hub.subscribe(subject, self, priority)

    def notify(self):
        pass

    def filter(self):
        return True


class SubHub(object):
    """
    订阅中心
    """
    def __init__(self, async_publish=None, logger=None):
        self.subject_map = {}
        self.subscriber_map = {}
        self.async_publish = async_publish
        self.logger = logger

    def set_logger(self, logger):
        self.logger = logger

    def set_async_publish(self, async_publish):
        self.async_publish = async_publish

    def register_subject(self, subject):
        if subject.name not in self.subject_map:
            self.subject_map[subject.name] = subject

    def create_or_get_subject(self, name):
        if name not in self.subject_map:
            self.subject_map[name] = Subject(name, self)
        return self.subject_map[name]

    def subscribe(self, subject, subscriber, priority=20):
        if isinstance(subject, Subject):
            name = subject.name
        else:
            name = subject
        if name not in self.subscriber_map:
            pqueue = []
            self.subscriber_map[name] = pqueue
        else:
            pqueue = self.subscriber_map[name]
        pqueue.append((priority, subscriber))
        pqueue.sort(key=lambda x: x[0])

    def subscribe_subjects(self, subscriber, subjects, priority=20):
        for s in subjects:
            self.subscribe(s, subscriber, priority)

    def connect_subscribers(self, subject, subscribers, priority=20):
        for s in subscribers:
            self.subscribe(subject, s, priority)

    def sync_publish(self, subject, *args, **kwargs):
        if isinstance(subject, Subject):
            name = subject.name
        else:
            name = subject
        subscribers = self.subscriber_map.get(name, [])
        context = kwargs.pop('context', {})
        kwargs['context'] = context
        for priority, subscriber in subscribers:
            try:
                s = subscriber(self, subject, *args, **kwargs)
                if s.filter():
                    s.notify()
                    if s.context:
                        context.update(s.context)
            except Exception as e:
                if self.logger:
                    self.logger.exception(e.message)
                    self.logger.error('subject {} subscriber {} args {} kwargs {}'.format(
                        subject,
                        subscriber.__name__,
                        args,
                        kwargs
                    ))

    def publish(self, subject, *args, **kwargs):
        """
        发布内容
        :param subject:
        :param args:
        :param kwargs:
        :return:
        """
        if isinstance(subject, Subject):
            name = subject.name
        else:
            name = subject
        publish_func = self.async_publish if self.async_publish else self.sync_publish
        publish_func(name, *args, **kwargs)


subscribe_hubs = {}
async_publish_funcs = {}
celery_task_of_hubs = {}


def create_async_publish_function(name, hub):
    """
    创建异步发布方法
    :param name:
    :param hub:
    :return:
    """
    @shared_task(name=name)
    def inner(subject, *args, **kwargs):
        hub.sync_publish(subject, *args, **kwargs)
    return inner.delay


def get_or_create_subscribe_hub(hub_name=None, celery_task=None, log_obj=None):
    """
    创建hub，或取得存在的Hub
    :param hub_name:
    :param celery_task:
    :param log_obj:
    :return:
    """
    if hub_name in subscribe_hubs:
        return subscribe_hubs[hub_name]
    _hub = SubHub()
    async_publish = None
    if celery_task:
        if celery_task not in async_publish_funcs:
            async_publish = create_async_publish_function(celery_task, _hub)
            async_publish_funcs[celery_task] = async_publish
            celery_task_of_hubs[_hub] = celery_task
        else:
            raise Exception('subscribe hub celery task {} had been used.'.format(celery_task))
    if not log_obj:
        log_obj = logging.getLogger(LOGGER_NAME)
    _hub.set_logger(log_obj)
    _hub.set_async_publish(async_publish)
    subscribe_hubs[hub_name] = _hub
    return _hub



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/29
 """
from __future__ import unicode_literals, absolute_import
from djdg_core.cache.redis_client import cache
from djdg_core.exceptions import TimeOutError
from djdg_core.cache.redlock import Redlock
import time


REDLOCK = Redlock([cache])


class CacheProxy(object):
    """
    缓存代理
    """
    FORMAT = None
    LOCK_PREFIX = 'lock'
    LOCK_EXPIRE = 300
    TIMESTAMP_RATE = 1
    KEY_EXPIRE = None
    WAIT_INTERVAL = 1

    def __init__(self, engine=None, redlock=None, **kwargs):
        if not engine:
            engine = cache
        if not redlock:
            redlock = REDLOCK
        self.logger = kwargs.pop('logger', None)
        self._engine = engine
        self._kwargs = kwargs
        self._key = None
        self._lock_key = None
        self._redlock = redlock
        self._lock = None

    def __getattr__(self, item):
        if item.startswith('hook_'):
            item = item[5:]

            def wrapper(*args, **kwargs):
                return self.hook(item, *args, **kwargs)
            return wrapper
        return getattr(self._engine, item)

    def hook(self, func, *args, **kwargs):
        """
        统一通过管道方式执行命令
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        expire = kwargs.pop('ex pire', None)
        pipe = self._engine.pipeline()
        call_func = getattr(pipe, func)
        pipe = call_func(self.key, *args, **kwargs)
        if expire and self.KEY_EXPIRE:
            pipe = pipe.expire(self.key, self.KEY_EXPIRE)
        r = pipe.execute()
        return r[0]

    @property
    def formatter(self):
        fmts = []
        for s in self.__class__.__mro__:
            fmt = getattr(s, 'FORMAT', None)
            if fmt:
                fmts.append(fmt)
        fmts.reverse()
        return ':'.join(fmts)

    @property
    def key(self):
        if not self._key:
            self._key = self.formatter.format(**self._kwargs)
        return self._key

    @property
    def lock_key(self):
        if not self._lock_key:
            self._lock_key = ':'.join([self.key, self.LOCK_PREFIX])
        return self._lock_key

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def acquire(self):
        start = time.time()
        while True:
            nt = time.time()
            lock = self._redlock.lock(self.lock_key, self.LOCK_EXPIRE * 1000)
            if lock:
                self._lock = lock
                break
            # 检查是否超时
            if nt - start > self.LOCK_EXPIRE:
                raise TimeOutError('获取锁{}超时！'.format(self.lock_key))
            time.sleep(self.WAIT_INTERVAL)
            if self.logger:
                self.logger.info('getting {} lock, wait {}'.format(self.key, time.time() - nt))

    def release(self):
        """
        acquire保证同时只有一个客户端获得锁，因此这里直接删除
        :return:
        """
        self._redlock.unlock(self._lock)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/8/8
 """
from __future__ import unicode_literals, absolute_import
from djdg_core.core_auth.middleware import AuthenticationMiddleware
from functools import update_wrapper
from djdg_core.exceptions import UserNotLoginError


def permit_all_access(view):
    setattr(view, AuthenticationMiddleware.PERMIT_ALL_PARAM, True)
    return view


def login_required(view):
    def wrapper(*args, **kwargs):
        if hasattr(args[0], 'request'):
            request = args[0].request
        else:
            request = args[0]
        if not request.user.is_authenticated():
            raise UserNotLoginError()
        return view(*args, **kwargs)
    return update_wrapper(wrapper, view)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2017/5/12
 """
from __future__ import unicode_literals, absolute_import
from django.db import models


class DateTimeModel(models.Model):
    """
    时间字段
    """
    ctime = models.DateTimeField('创建时间', auto_now_add=True)
    utime = models.DateTimeField('修改时间', auto_now=True)

    class Meta:
        abstract = True


def get_filter_manager(*args, **kwargs):
    """
    生成一个带过滤器的manager
    :param args:
    :param kwargs:
    :return:
    """
    class FilterManager(models.Manager):
        def get_queryset(self):
            return super(FilterManager, self).get_queryset().filter(*args, **kwargs)
    return FilterManager



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/12/21
 """
from __future__ import unicode_literals, absolute_import
from django import dispatch


register_event = dispatch.Signal(providing_args=['user'])

login_event = dispatch.Signal(providing_args=['user'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/20
 """
from __future__ import unicode_literals, absolute_import
from django import dispatch


# 用户通过验证码登录或注册事件
user_login_by_sms_code = dispatch.Signal(providing_args=['user'])

# 用户通过密码登录事件
user_login_by_password = dispatch.Signal(providing_args=['user'])

# 用户通过验证码与密码注册事件
user_register_by_password = dispatch.Signal(providing_args=['user'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/11/3
 """
from __future__ import unicode_literals, absolute_import
from django.db.transaction import atomic
from djdg_core.memberbao.public import memberbao_client
from djdg_core.memberbao.models import RefundRecord
from djdg_core.memberbao.models import PayResult
from djdg_core.exceptions import StanderResponseError
from djdg_core.memberbao import errors


def allocate(notify_id, app_id, operator_id, allocate_rules, request_no):
    """
    会员宝分账接口
    :param notify_id:
    :param app_id:
    :param operator_id:
    :param allocate_rules:
    :param request_no:
    :return:
    """
    res = memberbao_client.get_resource('order')
    data = {
        'requestNo': request_no,
        'notifyId': notify_id,
        'merchantId': app_id,
        'operatorId': operator_id,
        'allocateRules': allocate_rules
    }
    ret = res.post.allocate(data=data)
    return ret.data


def refund(amount, order_id, user_id):
    """
    会员宝退款
    :param amount:
    :param order_id:
    :param user_id:
    :return:
    """
    refund_record = None
    ret = None
    with atomic():
        pay_result = PayResult.objects.filter(order_id=order_id).first()
        if not pay_result:
            raise StanderResponseError(
                errors.PAY_RESULT_NOT_FOUND,
                '没有找到订单[{}]的支付结果记录！'.format(order_id)
            )
        res = memberbao_client.get_resource('refund')
        data = {
            'notifyId': pay_result.notify_id,
             'refundAmount': amount
        }
        ret = res.post(data=data)
        refund_record = RefundRecord.objects.create(
            notify_id=pay_result.notify_id,
            amount=amount,
            order_id=order_id,
            refund_id=ret.data['refundId'],
            user_id=user_id,
            confirm_amount=ret.data['refundAmount']
        )
    dif = abs(refund_record.amount - refund_record.confirm_amount)
    if dif > 0.001:
        raise StanderResponseError(
            errors.REFUND_AMOUNT_NOT_RIGHT,
            '会员宝退款金额[{am}/{cam}]不一致！'.format(
                am=refund_record.amount,
                cam=refund_record.confirm_amount

            )
        )
    return ret.data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2017/1/4
 """
from __future__ import unicode_literals, absolute_import
import jpush
from djdg_core.apis.jg.public import client
import logging
import json


logger = logging.getLogger('jpush')


def notification(alert, alias, platform=jpush.all_):
    """
    极光推送通知
    :param alert:
    :param alias: [phone number, ]
    :param platform:
    :return:
    """
    push = client.create_push()
    push.audience = jpush.audience(jpush.alias(alias))
    push.notification = jpush.notification(alert=alert)
    push.platform = platform
    res = push.send()
    logger.info('status code {}, response {}'.format(
        res.get_status_code(),
        json.dumps(res.payload)
    ))
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/19
 """
from __future__ import unicode_literals, absolute_import
from functools import update_wrapper


def decorate_urlpatterns(urlpatterns, decorator):
    for pattern in urlpatterns:
        if hasattr(pattern, 'url_patterns'):
            decorate_urlpatterns(pattern.url_patterns, decorator)

        if hasattr(pattern, '_callback'):
            pattern._callback = update_wrapper(decorator(pattern.callback), pattern.callback)

    return urlpatterns



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/18
 """
from __future__ import unicode_literals, absolute_import
from django.conf import settings
from django.utils.module_loading import import_string


RESPONSE_CONFIG = getattr(settings, 'RESPONSE_CONFIG')
AUTH_TOKEN_FORMAT = getattr(settings, 'AUTH_TOKEN_FORMAT')
MIDDLEWARE_CONFIG = getattr(settings, 'MIDDLEWARE_CONFIG')


core_settings = settings.DJDG_CORE

LOGGER_NAME = core_settings['LOGGER']
USER_PROXY = core_settings.get('USER_PROXY')
USER_MODEL = core_settings.get('USER_MODEL')
if USER_MODEL:
    USER_MODEL['MODEL'] = import_string(USER_MODEL['MODEL'])
PROJECT_NAME = core_settings.get('PROJECT_NAME', '')
EXT_USER = core_settings.get('EXT_USER', {})

if 'RESPONSE_CONFIG' in core_settings:
    RESPONSE_CONFIG = core_settings['RESPONSE_CONFIG']
if 'AUTH_TOKEN_FORMAT' in core_settings:
    AUTH_TOKEN_FORMAT = core_settings['AUTH_TOKEN_FORMAT']
if 'MIDDLEWARE_CONFIG' in core_settings:
    MIDDLEWARE_CONFIG = core_settings['MIDDLEWARE_CONFIG']


if 'USER_SYSTEM' in core_settings and 'COMMON_API' in core_settings:
    user_system_settings = core_settings['USER_SYSTEM']
    common_api_settings = core_settings['COMMON_API']
elif 'USER_SYSTEM' in core_settings:
    user_system_settings = common_api_settings = core_settings['USER_SYSTEM']
elif 'COMMON_API' in core_settings:
    user_system_settings = common_api_settings = core_settings['COMMON_API']
else:
    raise Exception('请填写用户系统和公共模块相关配置！')


cache_settings = core_settings.get('CACHE', {})
cache_token_settings = cache_settings.get('TOKEN', {})


USER_SYSTEM_HOST = user_system_settings['HOST']
USER_SYSTEM_BASE_URL = user_system_settings['BASE_URL']
USER_SYSTEM_PROTOCOL = user_system_settings['PROTOCOL']
USER_SYSTEM_AUTH_APP_ID = user_system_settings.get('APPID')
USER_SYSTEM_AUTH_APP_SECRET = user_system_settings.get('SECRET')
USER_SYSTEM_USER_RESOURCE = user_system_settings.get('USER_RESOURCE')


COMMON_API_HOST = common_api_settings['HOST']
COMMON_API_BASE_URL = common_api_settings['BASE_URL']
COMMON_API_PROTOCOL = common_api_settings['PROTOCOL']
COMMON_API_AUTH_APP_ID = common_api_settings['APPID']
COMMON_API_AUTH_APP_SECRET = common_api_settings['SECRET']


memberbao_settings = core_settings.get('MEMBERBAO', {})
MEMBERBAO_HOST = memberbao_settings.get('HOST')
MEMBERBAO_BASE_URL = memberbao_settings.get('BASE_URL')
MEMBERBAO_PROTOCOL = memberbao_settings.get('PROTOCOL')
MEMBERBAO_AUTH_APP_ID = memberbao_settings.get('APPID')
MEMBERBAO_AUTH_APP_SECRET = memberbao_settings.get('SECRET')


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/9/13
 """
from __future__ import unicode_literals, absolute_import
from rest_framework import serializers
from collections import OrderedDict
import datetime
import time


class ChoicesField(serializers.Field):
    """
    选择数据的字段
    """
    def __init__(self, choices, **kwargs):
        self._choices = OrderedDict(choices)
        self._rel_choices = OrderedDict([[v, k] for k, v in choices])
        super(ChoicesField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        if data in self._choices:
            return data
        v = self._rel_choices.get(data, None)
        if v is not None:
            return v
        msg_list = ['{0}: {1}'.format(v, k) for k, v in self._choices.items()]
        raise serializers.ValidationError("仅接受以下参数 {0}.".format(', '.join(msg_list)))

    def to_representation(self, value):
        return self._choices[value]


class TimestampField(serializers.DateTimeField):
    """
    时间戳，内部表示为datetime，序列化为int
    """
    def to_internal_value(self, data):
        if isinstance(data, int):
            data = datetime.datetime.fromtimestamp(data)
        return super(TimestampField, self).to_internal_value(data)

    def to_representation(self, value):
        return int(time.mktime(value.timetuple()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/10/18
 """
from __future__ import unicode_literals, absolute_import
from djdg_core.rest_client import RestClient
from djdg_core import settings
import logging


logger = logging.getLogger(settings.LOGGER_NAME)


common_client = RestClient(
    host=settings.COMMON_API_HOST,
    base_url=settings.COMMON_API_BASE_URL,
    protocol=settings.COMMON_API_PROTOCOL,
    logger=logger,
    oauth_app_id=settings.COMMON_API_AUTH_APP_ID,
    oauth_app_secret=settings.COMMON_API_AUTH_APP_SECRET
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/11/2
 """
from __future__ import unicode_literals, absolute_import
from djdg_core.exceptions import StanderResponseError
from djdg_core.common_api.public import common_client


def get_code(wx_redirect, appid):
    """
    发送回调url，以获取code
    :param wx_redirect: 微信回调url
    :param appid:
    :return:
    """
    res = common_client.get_resource('wx')
    res.post.code(data={
        'wxRedirect': wx_redirect,
        'appid': appid
    })


def get_access_token(code, appid, secret):
    """
    通过code获取access_token
    :param code:
    :param appid:
    :param secret:
    :return:
    """
    res = common_client.get_resource('wx')
    data = {
        'code': code,
        'appid': appid,
        'secret': secret
    }
    ret = res.post.oauth(data=data, raise_exception=False)
    result = ret.result
    if 'access_token' not in result:
        raise StanderResponseError(
            result.get('errcode', 500),
            result.get('errmsg', 'error')
        )
    return result['access_token']


def get_userinfo(openid, access_token):
    """
    获取用户信息
    :param openid:
    :param access_token:
    :return:
    """
    res = common_client.get_resource('wx')
    data = {
        'openid': openid,
        'access_token': access_token
    }
    ret = res.post.userinfo(data=data, raise_exception=False)
    result = ret.result
    if result.get('errmsg', None):
        raise StanderResponseError(
            result.get('errcode', 500),
            result.get('errmsg', 'error')
        )
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/9/18
 """
from __future__ import unicode_literals, absolute_import
from rest_framework import serializers
from djdg_core.exceptions import StandardResponseError
from collections import defaultdict
from rest_framework.serializers import html
from rest_framework.serializers import ValidationError
from rest_framework.serializers import api_settings
from rest_framework.serializers import empty
from django.db.models.fields.related import ManyToManyField
from rest_framework.serializers import LIST_SERIALIZER_KWARGS


def exclude_field(instance, field_name, include_fields=None, exclude_fields=None):
    f = instance._meta.get_field(field_name)
    if isinstance(f, ManyToManyField):
        return True
    elif include_fields and f.name not in include_fields:
        return True
    elif exclude_fields and f.name in exclude_fields:
        return True
    return False


class SimpleListSerializer(serializers.ListSerializer):
    """
    默认的ListSerializer update操作
    """
    def update(self, instance, validated_data):
        ist_map = {ist.id: ist for ist in instance}
        upd_set = set()
        ret = []
        for item in validated_data:
            if 'id' in item:
                upd_set.add(item['id'])
                # 更新现有数据
                ist = ist_map.get(item['id'], None)
                if ist:
                    ret.append(self.child.update(ist, item))
            else:
                # 创建新数据
                ret.append(self.child.create(item))
        # 删除数据
        ist_del = set(ist_map.keys()) - upd_set
        for ist_id in ist_del:
            ist = ist_map[ist_id]
            ist.delete()
        return ret


class DifferenceBaseSerializer(serializers.BaseSerializer):
    """
    差别化序列化
    """
    model_primary_key = 'id'
    include_fields = None
    exclude_fields = None

    def __init__(self, instance=None, data=empty, **kwargs):
        super(DifferenceBaseSerializer, self).__init__(instance=instance, data=data, **kwargs)
        self.difference = {}

    def _do_diff(self, validated_data, instance, diff_type):
        diff = {}
        serl_inst = getattr(self, 'child', self)
        if instance is not None:
            instance_data = serl_inst.to_representation(instance)
        else:
            instance_data = {}
        init_data = serl_inst.to_representation(validated_data)
        if 'create' == diff_type:
            for k, v in init_data.iteritems():
                if exclude_field(instance, k, self.include_fields, self.exclude_fields):
                    continue
                diff[k] = (None, v)
        elif 'update' == diff_type:
            for k, v in init_data.iteritems():
                if exclude_field(instance, k, self.include_fields, self.exclude_fields):
                    continue
                a = instance_data.get(k)
                if str(a) != str(v):
                    diff[k] = (a, v)
        elif 'delete' == diff_type:
            for k, v in instance_data.iteritems():
                if exclude_field(instance, k, self.include_fields, self.exclude_fields):
                    continue
                diff[k] = (v, None)
        else:
            raise StandardResponseError(7, '{}是无效的差异类型！'.format(diff_type))
        if diff:
            diff['__diff_type'] = diff_type
            diff[self.model_primary_key] = instance_data.get(self.model_primary_key)
            self._make_diff(diff)
            return True
        return False

    def _make_diff(self, diff):
        self.difference = diff


class DifferenceSerializer(DifferenceBaseSerializer, serializers.Serializer):
    """
    差别化序列化
    """
    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
            'child_cls': cls
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
                               key: value for key, value in kwargs.items()
                               if key in LIST_SERIALIZER_KWARGS
                               })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(meta, 'list_serializer_class', serializers.ListSerializer)
        return list_serializer_class(*args, **list_kwargs)

    def save(self, **kwargs):
        assert not hasattr(self, 'save_object'), (
            'Serializer `%s.%s` has old-style version 2 `.save_object()` '
            'that is no longer compatible with REST framework 3. '
            'Use the new-style `.create()` and `.update()` methods instead.' %
            (self.__class__.__module__, self.__class__.__name__)
        )

        assert hasattr(self, '_errors'), (
            'You must call `.is_valid()` before calling `.save()`.'
        )

        assert not self.errors, (
            'You cannot call `.save()` on a serializer with invalid data.'
        )

        # Guard against incorrect use of `serializer.save(commit=False)`
        assert 'commit' not in kwargs, (
            "'commit' is not a valid keyword argument to the 'save()' method. "
            "If you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
            "You can also pass additional keyword arguments to 'save()' if you "
            "need to set extra attributes on the saved model instance. "
            "For example: 'serializer.save(owner=request.user)'.'"
        )

        assert not hasattr(self, '_data'), (
            "You cannot call `.save()` after accessing `serializer.data`."
            "If you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
        )

        validated_data = dict(
            list(self.validated_data.items()) +
            list(kwargs.items())
        )
        if self.instance is not None:
            if self._do_diff(validated_data, self.instance, 'update'):
                self.instance = self.update(self.instance, validated_data)
                assert self.instance is not None, (
                    '`update()` did not return an object instance.'
                )
        else:
            self.instance = self.create(validated_data)
            self._do_diff(validated_data, self.instance, 'create')
            assert self.instance is not None, (
                '`create()` did not return an object instance.'
            )

        return self.instance


class DifferenceModelSerializer(DifferenceSerializer, serializers.ModelSerializer):
    """
    差别化序列化工具
    """


class DifferenceListSerializer(DifferenceBaseSerializer, serializers.ListSerializer):
    """
    差别化列表序列化工具
    """
    def __init__(self, *args, **kwargs):
        self.child_cls = kwargs.pop('child_cls', None)
        super(DifferenceListSerializer, self).__init__(*args, **kwargs)
        self.differences = defaultdict(list)
        self._separation = None
        self._validated_separation = defaultdict(list)

    def separation(self, data):
        if self._separation is None:
            instance_map = {getattr(ist, self.model_primary_key): ist for ist in self.instance} if self.instance is not None else {}
            _separation = defaultdict(list)
            keep_keys = set()
            for d in data:
                if self.model_primary_key not in d:
                    _separation['create'].append(d)
                elif d[self.model_primary_key] in instance_map:
                    _separation['update'].append([instance_map[d[self.model_primary_key]], d])
                    keep_keys.add(d[self.model_primary_key])
                else:
                    d.pop(self.model_primary_key, None)
                    _separation['create'].append(d)
            for k, i in instance_map.iteritems():
                if k not in keep_keys:
                    _separation['delete'].append(i)
            self._separation = _separation
        return self._separation

    def to_internal_value(self, data):
        """
        List of dicts of native values <- List of dicts of primitive datatypes.
        """
        if html.is_html_input(data):
            data = html.parse_html_list(data)

        if not isinstance(data, list):
            message = self.error_messages['not_a_list'].format(
                input_type=type(data).__name__
            )
            raise ValidationError({
                api_settings.NON_FIELD_ERRORS_KEY: [message]
            })

        if not self.allow_empty and len(data) == 0:
            message = self.error_messages['empty']
            raise ValidationError({
                api_settings.NON_FIELD_ERRORS_KEY: [message]
            })

        ret = []
        errors = []
        for k, vl in self.separation(data).iteritems():
            for v in vl:
                if 'create' == k:
                    init_kw = {'data': v}
                elif 'update' == k:
                    init_kw = {'data': v[1], 'instance': v[0]}
                else:
                    self._validated_separation[k].append(v)
                    continue
                try:
                    validated = self.child_cls(**init_kw).run_validation(init_kw['data'])
                    init_kw.pop('data', None)
                    sep_valid = dict(data=validated, **init_kw)
                except ValidationError as exc:
                    errors.append(exc.detail)
                else:
                    ret.append(validated)
                    self._validated_separation[k].append(sep_valid)
                    errors.append({})

        if any(errors):
            raise ValidationError(errors)

        return ret

    def save(self, **kwargs):
        """
        Save and return a list of object instances.
        """
        # Guard against incorrect use of `serializer.save(commit=False)`
        assert 'commit' not in kwargs, (
            "'commit' is not a valid keyword argument to the 'save()' method. "
            "If you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
            "You can also pass additional keyword arguments to 'save()' if you "
            "need to set extra attributes on the saved model instance. "
            "For example: 'serializer.save(owner=request.user)'.'"
        )
        instance_list = []
        for k, vl in self._validated_separation.iteritems():
            for v in vl:
                if 'create' == k:
                    attrs = dict(list(v['data'].items()) + list(kwargs.items()))
                    instance = self.child.create(attrs)
                    self._do_diff(attrs, instance, 'create')
                    instance_list.append(instance)
                elif 'update' == k:
                    attrs = dict(list(v['data'].items()) + list(kwargs.items()))
                    instance = v['instance']
                    if self._do_diff(attrs, v['instance'], 'update'):
                        instance = self.update(v['instance'], attrs)
                    instance_list.append(instance)
                else:
                    self._do_diff(None, v, 'delete')
                    self.delete(v)
        self.instance = instance_list
        return self.instance

    def _make_diff(self, diff):
        self.differences[diff['__diff_type']].append(diff)

    def delete(self, instance):
        instance.delete()
        return instance

    def update(self, instance, validated_data):
        for k, v in validated_data.iteritems():
            setattr(instance, k, v)
        instance.save()
        return instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/7/12
 """
from __future__ import unicode_literals, absolute_import
from django.conf import settings
from redis import StrictRedis
from djdg_core.exceptions import StandardResponseError
from redis.exceptions import RedisError
from redis._compat import iteritems
from redis.client import StrictPipeline
import re
import time


class MyStrictRedis(StrictRedis):
    """
    copy from o2o.cache to use REDIS.
    REDIS代理，将用到KEYS方法的KEY放入一个SET，使用KEYS操作时在此SET中查找
    """
    key_patterns = ('.*:column:map:set', '.*:groupon:set',
                    '.*:region:.*:promotion:set', 'seckill:.*',
                    '.*:seckill:set', 'smscode:.*', 'group:buying:*')

    def get_pattern_key(self, key):
        """
        """
        for pattern in self.key_patterns:
            if re.match(pattern, key):
                return pattern
        return None

    def keys(self, pattern='*'):
        pattern_key = self.get_pattern_key(pattern)
        if pattern_key is None:
            return ()

        result = []
        pattern = pattern.replace('*', '.*')
        keys = self.smembers(pattern_key)
        if keys is None:
            return ()
        for key in keys:
            if re.match(pattern, key):
                result.append(key)

        return result

    def add_key(self, key):
        pattern_key = self.get_pattern_key(key)
        if pattern_key is not None:
            super(MyStrictRedis, self).sadd(pattern_key, key)

    def remove_key(self, key):
        pattern_key = self.get_pattern_key(key)
        # print pattern_key,key
        if pattern_key is not None:
            self.srem(pattern_key, key)

    def brpoplpush(self, listname, back_list):
        darr = super(MyStrictRedis, self).brpop(listname)
        if darr is None:
            return None
        data = darr[1]
        super(MyStrictRedis, self).lpush(back_list, data)
        return data

    def execute_command(self, *args, **options):
        try:
            return super(MyStrictRedis, self).execute_command(*args, **options)
        except Exception as e:
            print e.message
            pass

    def clean_keys(self, pattern_key):
        if pattern_key in self.key_patterns:
            names = super(MyStrictRedis, self).smembers(pattern_key)
            if not names:
                names = set()
            names.add(pattern_key)
            super(MyStrictRedis, self).delete(*names)

    def delete(self, *names):
        for name in names:
            self.remove_key(name)
        return super(MyStrictRedis, self).delete(*names)

    def hset(self, name, key, value):
        self.add_key(name)
        return super(MyStrictRedis, self).hset(name, key, value)

    def set(self, name, value, ex=None, px=None, nx=False, xx=False):
        self.add_key(name)
        return super(MyStrictRedis, self).set(name, value, ex, px, nx, xx)

    def rpush(self, name, *values):
        self.add_key(name)
        return super(MyStrictRedis, self).rpush(name, *values)

    def lpush(self, name, *values):
        self.add_key(name)
        return super(MyStrictRedis, self).lpush(name, *values)

    def sadd(self, name, *values):
        # log.info(name)
        self.add_key(name)
        return super(MyStrictRedis, self).sadd(name, *values)


class LimitedRedisKeyManager(object):
    """
    redis method parameter parser
    """
    STRING_COMMANDS = 'APPEND BITCOUNT BITOP DECR DECRBY GET GETBIT GETRANGE GETSET INCR INCRBY INCRBYFLOAT MGET ' \
                      'MSET MSETNX PSETEX SET SETBIT SETEX SETNX SETRANGE STRLEN'
    STRING_GET_VALUE_COMMANDS = 'BITCOUNT GET GETBIT GETRANGE MGET STRLEN'

    HASH_COMMANDS = 'HDEL HEXISTS HGET HGETALL HINCRBY HINCRBYFLOAT HKEYS HLEN HMGET HMSET HSET HSETNX HVALS HSCAN'
    HASH_GET_VALUE_COMMANDS = 'HEXISTS HGET HGETALL HKEYS HLEN HMGET HVALS HSCAN'

    LIST_COMMANDS = 'BLPOP BRPOP BRPOPLPUSH LINDEX LINSERT LLEN LPOP LPUSH LPUSHX LRANGE LREM LSET LTRIM RPOP ' \
                    'RPOPLPUSH RPUSH RPUSHX'
    LIST_GET_VALUE_COMMANDS = 'LINDEX LINSERT LLEN LRANGE LPUSHX LSET LTRIM RPUSHX'

    SET_COMMANDS = 'SADD SCARD SDIFF SDIFFSTORE SINTER SINTERSTORE SISMEMBER SMEMBERS SMOVE SPOP SRANDMEMBER SREM ' \
                   'SUNION SUNIONSTORE SSCAN'
    SET_GET_VALUE_COMMANDS = 'SCARD SDIFF SINTER SISMEMBER SMEMBERS SRANDMEMBER SUNION SSCAN'

    SORTED_SET_COMMANDS = 'ZADD ZCARD ZCOUNT ZINCRBY ZRANGE ZRANGEBYSCORE ZRANK ZREM ZREMRANGEBYRANK ' \
                         'ZREMRANGEBYSCORE ZREVRANGE ZREVRANGEBYSCORE ZREVRANK ZSCORE ZUNIONSTORE ZINTERSTORE ZSCAN'
    SORTED_SET_GET_VALUE_COMMANDS = 'ZCARD ZCOUNT ZRANGE ZRANGEBYSCORE ZRANK ZREVRANGE ZREVRANGEBYSCORE ZREVRANK ' \
                                    'ZSCORE ZSCAN'

    DEL_COMMAND = 'DEL'

    MAYBE_DEL_COMMANDS = 'HDEL ' \
                         'BLPOP BRPOP LPOP LREM RPOP ' \
                         'SPOP SREM ' \
                         'ZREM'

    REDIS_DATA_TYPES = ['string', 'hash', 'list', 'set', 'sorted_set']

    def __init__(self, redis):
        self.redis = redis
        self.command_group_map = {}
        for s, t in zip([self.STRING_COMMANDS, self.HASH_COMMANDS, self.LIST_COMMANDS, self.SET_COMMANDS,
                         self.SORTED_SET_COMMANDS],
                        self.REDIS_DATA_TYPES):
            for c in s.split(' '):
                self.command_group_map[c] = t
        self.get_value_commands = {}
        for s, t in zip([self.STRING_GET_VALUE_COMMANDS, self.HASH_GET_VALUE_COMMANDS, self.LIST_GET_VALUE_COMMANDS,
                         self.SET_GET_VALUE_COMMANDS, self.SORTED_SET_GET_VALUE_COMMANDS], self.REDIS_DATA_TYPES):
            for c in s.split(' '):
                self.get_value_commands[c] = t
        self.maybe_del_commands = {c for c in self.MAYBE_DEL_COMMANDS.split(' ')}

    def remove_from_container(self, *keys):
        if not keys:
            return None
        return self.redis.rem_keys(*keys)

    def put_in_container(self, *keys):
        if not keys:
            return None
        ts = time.time()
        kwargs = {k: ts for k in keys}
        return self.redis.add_keys(**kwargs)

    def detect_and_remove_keys(self, *keys):
        # pipe = self.redis.pipeline()
        # for k in keys:
        #     pipe = pipe.exists(k)
        res = [self.redis.exists(k) for k in keys]
        removable_keys = [k for k, e in zip(keys, res) if not e]
        self.remove_from_container(*removable_keys)

    def detect_and_add_keys(self, *keys):
        res = [self.redis.exists(k) for k in keys]
        add_keys = [k for k, e in zip(keys, res) if e]
        self.put_in_container(*add_keys)

    def manage_keys(self, response, *args):
        cmd, keys, cmd_type = self.extract_command_keys(*args)
        # 只读，不产生新key，跳过
        if not keys:
            return None
        if self.DEL_COMMAND == cmd:
            self.remove_from_container(*keys)
        elif cmd in self.maybe_del_commands:
            self.detect_and_remove_keys(*keys)
        else:
            m = getattr(self, 'manage_{}_keys'.format(cmd_type))
            m(cmd, keys)

    def manage_string_keys(self, cmd, keys):
        self.put_in_container(*keys)

    def manage_hash_keys(self, cmd, keys):
        self.put_in_container(*keys)

    def manage_list_keys(self, cmd, keys):
        if cmd.endswith('RPOPLPUSH'):
            self.detect_and_remove_keys(*[keys[0]])
            self.detect_and_add_keys(*[keys[1]])
        else:
            self.put_in_container(*keys)

    def manage_set_keys(self, cmd, keys):
        self.put_in_container(*keys)

    def manage_sorted_set_keys(self, cmd, keys):
        self.put_in_container(*keys)

    def extract_command_keys(self, *args):
        cmd = args[0]
        # 如果是仅仅获取值，这种是不会产生新的key，直接返回
        if cmd in self.get_value_commands:
            return cmd, [], None
        if self.DEL_COMMAND == cmd:
            return cmd, args[1:], 'key'
        cmd_type = self.command_group_map.get(cmd)
        if not cmd_type:
            return cmd, [], None
        m = getattr(self, 'extract_{}_keys'.format(cmd_type), None)
        if not m:
            return cmd, [], cmd_type
        return cmd, m(cmd, args[1:]), cmd_type

    @classmethod
    def extract_string_keys(cls, cmd, params):
        keys = []
        if 'BITOP' == cmd:
            keys.append(params[2])
        elif cmd.startswith('MSET'):
            keys = [e for i, e in enumerate(params) if not (i % 2)]
        else:
            keys.append(params[0])
        return keys

    @classmethod
    def extract_hash_keys(cls, cmd, params):
        return [params[0]]

    @classmethod
    def extract_list_keys(cls, cmd, params):
        if cmd in ('BLPOP', 'BRPOP'):
            keys = params[:-1]
        elif cmd.endswith('RPOPLPUSH'):
            keys = params[:2]
        else:
            keys = [params[0]]
        return keys

    @classmethod
    def extract_set_keys(cls, cmd, params):
        if cmd in ('SDIFFSTORE', 'SINTERSTORE', 'SUNIONSTORE'):
            keys = params[:]
        elif 'SMOVE' == cmd:
            keys = params[:2]
        else:
            keys = [params[0]]
        return keys

    @classmethod
    def extract_sorted_set_keys(cls, cmd, params):
        keys = [params[0]]
        return keys


class LimitedStrictRedisPipeline(StrictPipeline):
    """
    对于管道模式，照样要把key纳入管理
    """
    def __init__(self, *args, **kwargs):
        self.key_manager = kwargs.pop('key_manager', None)
        super(LimitedStrictRedisPipeline, self).__init__(*args, **kwargs)

    def execute(self, raise_on_error=True, managed=True):
        """
        :param raise_on_error:
        :param managed:
        :return:
        """
        managed_commands = []
        for args, options in self.command_stack:
            _m = options.pop('managed', True)
            if _m and managed:
                managed_commands.append((args, True))
            else:
                managed_commands.append((args, False))
        res = super(LimitedStrictRedisPipeline, self).execute(raise_on_error)
        for r, c in zip(res, managed_commands):
            if not c[1]:
                continue
            self.key_manager.manage_keys(r, c[0])
        return res

    def immediate_execute_command(self, *args, **options):
        managed = options.pop('managed', True)
        res = super(LimitedStrictRedisPipeline, self).immediate_execute_command(*args, **options)
        if managed:
            self.key_manager.manage_keys(res, args)
        return res


class LimitedStrictRedis(StrictRedis):
    """
    受限的redis客户端
    """
    KEYS_CONTAINER = 'limited:redis:keys'

    def __init__(self, *args, **kwargs):
        super(LimitedStrictRedis, self).__init__(*args, **kwargs)
        self.key_manager = LimitedRedisKeyManager(self)

    def execute_command(self, *args, **options):
        """
        重载 execute_command 将操作的key都存储到KEYS_CONTAINER里面
        :param args:
        :param options:
        :return:
        """
        managed = options.pop('managed', True)
        cmd = args[0]
        if not managed:
            return super(LimitedStrictRedis, self).execute_command(*args, **options)
        elif cmd == self.KEYS_CONTAINER:
            raise StandardResponseError(1, '被管理的key不能是{}'.format(self.KEYS_CONTAINER))
        res = super(LimitedStrictRedis, self).execute_command(*args, **options)
        self.key_manager.manage_keys(res, *args)
        return res

    def pipeline(self, transaction=True, shard_hint=None):
        return LimitedStrictRedisPipeline(
            self.connection_pool,
            self.response_callbacks,
            transaction,
            shard_hint,
            key_manager=self.key_manager
        )

    def exists(self, name):
        return super(LimitedStrictRedis, self).execute_command('EXISTS', name)

    __contains__ = exists

    def add_keys(self, *args, **kwargs):
        """
        往key容器里面增加键
        """
        pieces = []
        if args:
            if len(args) % 2 != 0:
                raise RedisError("ZADD requires an equal number of "
                                 "values and scores")
            pieces.extend(args)
        for pair in iteritems(kwargs):
            pieces.append(pair[1])
            pieces.append(pair[0])
        return super(LimitedStrictRedis, self).execute_command('ZADD', self.KEYS_CONTAINER, *pieces)

    def rem_keys(self, *values):
        """
        从key容器里面移除键
        :param values:
        :return:
        """
        return super(LimitedStrictRedis, self).execute_command('ZREM', self.KEYS_CONTAINER, *values)

    def bitop(self, operation, dest, *keys):
        raise StandardResponseError(1, 'bitop 该命令不支持！')

    def msetnx(self, *args, **kwargs):
        raise StandardResponseError(1, 'msetnx 该命令不支持！')

    def blpop(self, keys, timeout=0):
        raise StandardResponseError(1, 'blpop 该命令不支持！')

    def brpop(self, keys, timeout=0):
        raise StandardResponseError(1, 'brpop 该命令不支持！')

    def brpoplpush(self, src, dst, timeout=0):
        raise StandardResponseError(1, 'brpoplpush 该命令不支持！')


class DefaultCacheProxy(object):
    """
    Proxy access to the default Cache object's attributes.

    This allows the legacy `cache` object to be thread-safe using the new
    ``caches`` API.
    """
    def __init__(self):
        self._client = None

    @property
    def client(self):
        if not self._client:
            kwargs = {}
            for k, v in settings.REDIS.items():
                kwargs[k.lower()] = v
            # self._client = MyStrictRedis(**kwargs)
            # self._client = LimitedStrictRedis(**kwargs)
            self._client = StrictRedis(**kwargs)
        return self._client

    def __getattr__(self, name):
        return getattr(self.client, name)


cache = DefaultCacheProxy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/8/6
 """
from __future__ import unicode_literals, absolute_import
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User as AuthUser
from djdg_core.core_auth.models import OpenUser
from djdg_oauth.oauthcore import getSign
from djdg_core.core_auth.token import get_user_id
from djdg_core.core_auth.token import get_user_by_token


class TokenAuthenticationBackend(ModelBackend):
    """
    token方式验证逻辑
    """
    def authenticate(self, token):
        user = get_user_by_token(token)
        if user.is_authenticated():
            return user
        return None


class OpenAuthenticationBackend(ModelBackend):
    """
    oauth方式认证
    """
    def authenticate(self, body, appid, secret, signature):
        _signature = getSign(body, secret)
        if _signature != signature:
            return None
        user = OpenUser(appid, secret)
        return user


class AuthUserTokenAuthenticationBackend(ModelBackend):
    """
    token方式验证逻辑
    """
    def authenticate(self, token):
        user_id = get_user_id(token)
        if not user_id:
            return None
        user = AuthUser.objects.filter(pk=user_id).first()
        return user



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/11/14
 """
from __future__ import unicode_literals, absolute_import


PAY_RESULT_NOT_FOUND = 6000
REFUND_AMOUNT_NOT_RIGHT = 6001


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/8/5
 """
from __future__ import unicode_literals, absolute_import
from django.utils.translation import ugettext as _


class StanderResponseError(Exception):
    """
    标准响应错误，必须包含 status 和 msg 两个字段
    data 为在发生错误的情况下需要给前端返回的数据，一般是没有的。

    0-9999 公共错误代码
    10000-19999 o2o
    20000-29999 商家版
    30000-39999 配送版
    40000-49999 供应链
    """
    def __init__(self, status, msg, data=None):
        self.status = status
        self.msg = msg
        self.data = data
        self.message = msg


auth_error = StanderResponseError(1, _('not authenticated'), {})


class StandardResponseError(StanderResponseError):
    """
    纠正拼写
    """


class ParameterNotExistsError(StanderResponseError):
    """
    参数不存在
    """
    def __init__(self, param):
        msg = _('Parameter {0} dose not exists.').format(param)
        super(ParameterNotExistsError, self).__init__(2, msg)


class ParameterNotExistOrEmptyError(ParameterNotExistsError):
    """
    参数为空或者不存在
    """

    def __init__(self, param, msg=None):
        if msg is None:
            msg = _('Parameter {0} dose not exists or empty.').format(param)
        super(ParameterNotExistsError, self).__init__(2, msg)


rest_error = StanderResponseError(3, _('rest framework error'))

signature_error = StanderResponseError(4, _('signature is not right'))

state_error = StanderResponseError(5, _('state error'))


class ParameterNotPositiveError(StanderResponseError):
    """
    非整数错误
    """
    def __init__(self, param):
        msg = '参数[{}]应该是一个正数！'.format(param)
        super(ParameterNotPositiveError, self).__init__(6, msg)


class UserNotExists(StanderResponseError):
    """
    用户不存在
    """
    def __init__(self, username):
        msg = '用户[{}]不存在！'.format(username)
        super(UserNotExists, self).__init__(7, msg)


class PasswordError(StanderResponseError):
    """
    密码不正确
    """
    def __init__(self, username):
        msg = '用户[{}]的密码不正确！'.format(username)
        super(PasswordError, self).__init__(8, msg)


class TimeOutError(StanderResponseError):
    """
    超时
    """
    def __init__(self, msg, data=None):
        super(TimeOutError, self).__init__(9, msg, data)


class DifferentPasswordError(StanderResponseError):
    """
    两次密码输入不一致
    """
    def __init__(self, data=None):
        super(DifferentPasswordError, self).__init__(10, '两次密码输入不一致!', data)


class CodeVerifyError(StanderResponseError):
    """
    验证码不正确！
    """
    def __init__(self, status=11, data=None):
        super(CodeVerifyError, self).__init__(status, '验证码不正确！', data)


class UserNameExistsError(StanderResponseError):
    """
    用户已经存在！
    """
    def __init__(self, status=12, msg=None, data=None):
        msg = msg if msg else '用户已经存在！'
        super(UserNameExistsError, self).__init__(status, msg, data)


class PermissionError(StanderResponseError):
    """
    权限错误
    """
    def __init__(self, status=13, msg=None, data=None):
        msg = msg if msg else '用户没有权限进行此操作！'
        super(PermissionError, self).__init__(status, msg, data)


class UserNotLoginError(StanderResponseError):
    """
    用户尚未登录
    """
    def __init__(self, status=14, msg=None, data=None):
        msg = msg if msg else '请先登录！'
        super(UserNotLoginError, self).__init__(status, msg, data)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
author lilx
created on 2016/9/20
 """
from __future__ import unicode_literals, absolute_import
from django.utils import six
from djdg_core.exceptions import StanderResponseError
from djdg_core.exceptions import state_error


class StateMetaclass(type):
    """
    状态元类
    """
    @classmethod
    def get_state_machine(mcs, bases, attrs):
        machine_class = attrs.get('machine_class')
        if not machine_class:
            for base in bases:
                machine_class = getattr(base, 'machine_class', None)
                if machine_class:
                    break
        return machine_class

    def __new__(mcs, name, bases, attrs):
        machine_class = mcs.get_state_machine(bases, attrs)
        cls = super(StateMetaclass, mcs).__new__(mcs, name, bases, attrs)
        if machine_class:
            machine_class.register(cls)
        return cls


@six.add_metaclass(StateMetaclass)
class State(object):
    """
    状态
    """
    value = None
    max_value = None
    machine_class = None

    def __init__(self, context):
        self.context = context

    def __cmp__(self, other):
        return cmp(self.value, other)

    def transition(self, state=None):
        if self.value >= self.max_value:
            raise StanderResponseError(state_error.status, '该状态不能往下迁移！')
        state_machine = self.context['state_machine']
        cur_val = state if state is not None else self.value + 1
        value = min(cur_val, self.max_value)
        next_status = state_machine.get_state(value)
        return next_status

    def is_next_state(self, state):
        next_state = self.transition()
        return next_state == state


class StateMachine(object):
    """
    状态机
    """
    _states = None
    current_state = None
    context = None

    def __init__(self, current_state, **kwargs):
        kwargs['state_machine'] = self
        self.context = kwargs
        if isinstance(current_state, State):
            current_state = current_state.value
        status_cls = self._states[current_state]
        self.current_state = status_cls(context=self.context)

    @classmethod
    def register(cls, state_cls):
        if state_cls.value is not None:
            cls._states[state_cls.value] = state_cls

    def set_state(self, state):
        next_state = self.current_state.transition(state)
        if next_state == state:
            self.current_state = next_state
            return True
        return False

    def get_state(self, state):
        return self._states[state](self.context)


def state_machine_class():
    return type(b'StateMachineTemporary', (StateMachine, ), {'_states': {}})


def state_class(base, **kwargs):
    return type(b'StateTemporary', (base, ), kwargs)
